/*
 * srr_sys.h: implementation of the system calls
 *
 *  Linux Loadable Module Implementation of QNX Send/Receive/Reply
 *	message passing, and related function.
 *
 *	QNX is a trademark of QSSL, Ltd. <www.qnx.com>
 *
 * Copyright (C) 2000 Cogent Real-Time Systems Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software in a file called 'COPYING'; if not, write to:
 * 	Free Software Foundation, Inc.,
 * 	59 Temple Place, Suite 330,
 * 	Boston, MA 
 * 	02111-1307  USA
 *
 * $Id: srr_sys.h 3613 2009-02-01 19:58:01Z andrew $
 */

#ifndef SRR_SYS_H
#define SRR_SYS_H

#include <linux/types.h>
#include <linux/rtmutex.h>
#include <linux/sched.h>
#include <linux/spinlock.h>

#include "srr_taskLocks.h"

/* Not used if TASK_SPECIFIC_LOCKS defined */
#define USE_SPIN_LOCKS	1
#define SPIN_LOCK_USE_MUTEX
//#undef SPIN_LOCK_USE_MUTEX
#define USE_RT_MUTEX

/*
 * MACROS for implementing a spin lock.  We are not allowed to use
 * functions.
 */
extern spinlock_t rcv_q_lock;
#ifndef TASK_SPECIFIC_LOCKS
#ifndef USE_RT_MUTEX
extern struct semaphore srr_module_mutex;
#else
extern struct rt_mutex srr_module_mutex;
#endif
#else /* TASK_SPECIFIC_LOCKS */
extern atomic_t map_readers;
extern atomic_t map_writers;
#endif

#define Srr_UseSpinLock()	SPIN_FLAG_TYPE _spinflags __attribute__((unused)) = 0

#ifdef USE_SPIN_LOCKS
#ifndef SPIN_LOCK_USE_MUTEX
#define Srr_SpinLock()		spin_lock_irqsave(&rcv_q_lock, _spinflags)
#define Srr_SpinLockInterruptible()		spin_lock_irqsave(&rcv_q_lock, _spinflags)
#define Srr_SpinUnlock()	spin_unlock_irqrestore(&rcv_q_lock, _spinflags)
//#define Srr_SpinLock()		spin_lock_irq(&rcv_q_lock)
//#define Srr_SpinUnlock()	spin_unlock_irq(&rcv_q_lock)

/* This is the lock that we should look at using if we do SMP.  Until
   we have spent the time to get this right, use a mutex?  We need
   something, or else we run into issues with task tables changing
   while we are swapping in user memory.  */
//#define Srr_SpinLock()					spin_lock(&rcv_q_lock)
//#define Srr_SpinLockInterruptible()		spin_lock(&rcv_q_lock)
//#define Srr_SpinUnlock()				spin_unlock(&rcv_q_lock)
#else
/* This replaces spin locks with a single mutex */
#ifndef USE_RT_MUTEX
#define Srr_SpinLock()					down (&srr_module_mutex)
#define Srr_SpinLockInterruptible()		down_interruptible (&srr_module_mutex)
#define Srr_SpinUnlock()				up (&srr_module_mutex)
#else
#define Srr_SpinLock()		rt_mutex_lock(&srr_module_mutex)
#define Srr_SpinLockInterruptible() rt_mutex_lock_interruptible(&srr_module_mutex, 1)
#define Srr_SpinUnlock()		rt_mutex_unlock(&srr_module_mutex)
#endif
#endif // SPIN_LOCK_USE_MUTEX
#else
/* This disables spin locks and lets the module run freely */
#define Srr_SpinLock()
#define Srr_SpinLockInterruptible()
#define Srr_SpinUnlock()
#endif // USE_SPIN_LOCKS

#include "srr_tt.h"

void SrrTaskWake(SrrTask* task, int state, const char* whence);
int SrrWakeFromPoll(SrrTask* task, int state);
// void SrrTaskSleep(SrrTask* task, int state);

int	SrrDoSend(SrrTm* map, SrrTask* task, pfnSrrBufRead rdfn, pfnSrrBufWrite wrfn);
int SrrModSend(struct SrrTm* map, struct SrrTask* task);
/* Michael Eaton - 2011/06/28 */
int SrrModSendTimeout(struct SrrTm* map, struct SrrTask* task);
int SrrDoSendTimeout(SrrTm* map, SrrTask* task, pfnSrrBufRead rdfn, pfnSrrBufWrite wrfn);

int SrrRecv(struct SrrTm* map, struct SrrTask* task);
/* Michael Eaton - 2011/06/28 */
int SrrRecvTimeout(struct SrrTm* map, struct SrrTask* task);
int SrrModReply(struct SrrTm* map, struct SrrTask* task);
/* Len Meakin - Len Meakin */
int SrrModRelay(struct SrrTm* map, struct SrrTask* task);
/* 2000/02/17 - Alessandro Sala */
int SrrModReadmsg(struct SrrTm* map, struct SrrTask* task);
/* 2000/03/27 - Alessandro Sala */
int SrrModWritemsg(struct SrrTm* map, struct SrrTask* task);

/*
** Begin MX
** 2000/10/15 John Green
*/
int SrrModSendmx(struct SrrTm* map, struct SrrTask* task);
/* Michael Eaton -- 2011/06/30 */
int SrrModSendmxTimeout(struct SrrTm* map, struct SrrTask* task);
int SrrRecvmx(struct SrrTm* map, struct SrrTask* task);
/* Michael Eaton -- 2011/06/30 */
int SrrRecvmxTimeout(struct SrrTm* map, struct SrrTask* task);
int SrrModReplymx(struct SrrTm* map, struct SrrTask* task);
int SrrModReadmsgmx(struct SrrTm* map, struct SrrTask* task);
int SrrModWritemsgmx(struct SrrTm* map, struct SrrTask* task);
/*
** End MX
*/

int SrrModNameAttach(struct SrrTm* map, struct SrrTask* task);
int SrrModNameDetach(struct SrrTm* map, struct SrrTask* task);
int SrrModNameLocate(struct SrrTm* map, struct SrrTask* task);

int SrrModProxyAttach(struct SrrTm* map, struct SrrTask* task);
int SrrModProxyDetach(struct SrrTm* map, struct SrrTask* task);
int SrrModTrigger(struct SrrTm* map, struct SrrTask* task);

int SrrDoTrigger (struct SrrTm* map, pid_t pid, int have_lock);

int SrrPFlags(struct SrrTm* map, struct SrrTask* task);

int SrrDumpState(struct SrrTm* map, char*, size_t, int* totlen);
int SrrDumpSin(SrrTm* map, char* buf, size_t size, int* totlen);
int SrrDumpProxy (SrrTm* map, char* buf, size_t size, int* totlen);
int SrrDumpIrq (SrrTm* map, char* buf, size_t size, int* totlen);
int SrrDumpVersion (SrrTm* map, char* buf, size_t size, int* totlen);
int SrrDumpNames(SrrTm* map, char* buf, size_t size, int* totlen);
int SrrDumpTimers (SrrTm* map, char* buf, size_t size, int *totlen);
int SrrDumpTaskBlocks (SrrTm* map, char* buf, size_t size, int *totlen);

#ifndef NO_IRQ
int SrrIrq_DoIRQCommand (SrrTm* map, struct SrrTask* task);
int SrrIrq_Detach (SrrTask* task, int handle);
void SrrIrq_DetachAll (SrrTask* task);
#endif

void SrrClearPendingSignals (struct task_struct *task);

#endif
