/*
 * srr_time.c: implementation of POSIX timers w/ QNX extensions
 *
 *  Linux Loadable Module Implementation of QNX Send/Receive/Reply
 *	message passing, and related functions.
 *
 *	QNX is a trademark of QSSL, Ltd. <www.qnx.com>
 *
 *	Copyright (C) 1999  Andrew Thomas <andrew@cogent.ca>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software in a file called 'COPYING'; if not, write to:
 * 	Free Software Foundation, Inc.,
 * 	59 Temple Place, Suite 330,
 * 	Boston, MA 
 * 	02111-1307  USA
 *
 * $Id: srr_time.c 324 2014-02-23 05:30:32Z mstormer $
 */

#include "module_versions.h"
#include "sysdep.h"

#include "srr_mod.h"
#include "srr_util.h"
#include "srr_tt.h"
#include "srr_time.h"
#include "srr_sys.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#include <linux/tqueue.h>
#endif
#include <linux/interrupt.h>

#ifndef NSEC_PER_SEC
#define NSEC_PER_SEC 1000000000
#endif

#ifdef CONFIG_UTIME
#include <linux/mutime.h>
#define NSEC_PER_JIFFY (NSEC_PER_SEC / HZ)
#ifndef NSEC_PER_USEC
#define NSEC_PER_USEC 1000L
#endif
#else
#ifndef USEC_PER_SEC
#define USEC_PER_SEC    1000000
#endif
#endif

int SrrTimerCount = 0;

void SrrPt_TimerExpiry(unsigned long tid);

#ifdef CONFIG_UTIME
static void ts_to_jiffies_and_usec(struct timespec *value,
		unsigned long *jiffies_ptr,
		unsigned long *usec_ptr)
{
	unsigned long sec = (unsigned) value->tv_sec;
	unsigned long nsec = (unsigned) value->tv_nsec;
	unsigned long jiffies_from_ns = nsec / NSEC_PER_JIFFY;

	if(sec>(ULONG_MAX - jiffies_from_ns)/HZ) {
		*jiffies_ptr = ULONG_MAX;
		*usec_ptr = 0;
		return;
	}

	*jiffies_ptr=(sec*HZ) + jiffies_from_ns;
	*usec_ptr=(nsec % NSEC_PER_JIFFY)/NSEC_PER_USEC;

}

static void jiffies_and_usec_to_ts(unsigned long jif, unsigned long usc,
		struct timespec *value){

	value->tv_sec=(long)(jif/HZ);
	value->tv_nsec=(long)((jif%HZ) * NSEC_PER_JIFFY);

	value->tv_sec += (long)(usc/USEC_PER_SEC);
	value->tv_nsec +=(long)((usc% USEC_PER_SEC) * NSEC_PER_USEC);

	if(value->tv_nsec > NSEC_PER_SEC) {
		value->tv_sec += value->tv_nsec / NSEC_PER_SEC;
		value->tv_nsec %= NSEC_PER_SEC;
	}
}
#endif /* CONFIG_UTIME */

static inline
	void srr_timeval_to_timespec(struct timespec* ts, struct timeval* tv)
{
	ts->tv_sec	= tv->tv_sec;
	ts->tv_nsec	= tv->tv_usec * 1000;
}

static inline
	void itimerval_to_itimerspec(struct itimerspec* its, struct itimerval *itv)
{
	srr_timeval_to_timespec(&its->it_interval, &itv->it_interval);
	srr_timeval_to_timespec(&its->it_value, &itv->it_value);
}

static inline
	void srr_timespec_sub(struct timespec* r, struct timespec* lhs,struct timespec* rhs)
{
	/* Normalize the right-hand side */
	while (rhs->tv_nsec > NSEC_PER_SEC)
	{
		rhs->tv_nsec -= NSEC_PER_SEC;
		rhs->tv_sec += 1;
	}
	/* Normalize the left-hand side */
	if (lhs->tv_nsec < rhs->tv_nsec)
	{
		lhs->tv_sec -= 1;
		lhs->tv_nsec += NSEC_PER_SEC;
	}
	/* If the result will be negative, set it -1 nanosecond.  We do
	   not need accuracy, we need to have time >=0 for all combinations
	   of signed/unsigned in tv_sec and tv_nsec.  We will test for
	   -1 nanosecond in a special case below. */
	if (lhs->tv_sec < rhs->tv_sec ||
		(lhs->tv_sec == rhs->tv_sec &&
		 lhs->tv_nsec < rhs->tv_nsec))
	{
		r->tv_sec = 0;
		r->tv_nsec = (long)-1;
	}
	else
	{
		r->tv_sec  = lhs->tv_sec  - rhs->tv_sec;
		r->tv_nsec = lhs->tv_nsec - rhs->tv_nsec;
	}
}

static inline int itimerspec_isset(struct itimerspec* its)
{
	if (its->it_value.tv_sec || its->it_value.tv_nsec)
		return 1;
	else
		return 0;
}

/* atomic_add_unless was introduced sometime in the 2.6.x kernel series
   after 2.6.9.  This is the code from kernel 2.6.22. */
#ifndef atomic_cmpxchg
#define atomic_cmpxchg(v,old,new) (cmpxchg(&((v)->counter),(old),(new)))

#if !defined(_ASM_X86_ATOMIC_64_H) && !defined(_ASM_X86_ATOMIC_32_H) && !defined(_ASM_X86_ATOMIC_H) 
static __inline__ int atomic_add_unless(atomic_t *v, int a, int u)
{
	int c, old;
	c = atomic_read(v);
	for (;;) {
		if (unlikely(c == (u)))
			break;
		old = atomic_cmpxchg((v), c, c + (a));
		if (likely(old == c))
			break;
		c = old;
	}
	return c != (u);
}
#endif // _ASM_X86_ATOMIC_64_H
#endif // atomic_cmpxchg

extern SrrTm* srr_taskmap; /* this is a hack, but necessary for now */

SrrPt srr_pt[NR_TIMERS];

spinlock_t srr_module_timer_lock = SPIN_LOCK_UNLOCKED;

#define USES_TIMER_LOCK()	unsigned long _lock_flags
#define TIMER_LOCK()	spin_lock_irqsave(&srr_module_timer_lock, _lock_flags)
#define TIMER_UNLOCK()	spin_unlock_irqrestore(&srr_module_timer_lock, _lock_flags)
//#define TIMER_LOCK()
//#define TIMER_UNLOCK()

// A timer can be used (assigned to a task), scheduled (bottom half will run)
// or both.
// FREE = available for reuse
// USED = assigned to a task
// SCHEDULED = bottom half will run.  Maybe this can be scheduled on each CPU?

#define SET_UNUSED(tid) (atomic_set(&srr_pt[tid].is_used, 0))
#define SET_USED(tid) (atomic_set(&srr_pt[tid].is_used, 1))
#define IS_FREE(tid) (atomic_read(&srr_pt[tid].is_scheduled) == 0 && atomic_read(&srr_pt[tid].is_used) == 0)
#define IS_USED(tid) (atomic_read(&srr_pt[tid].is_used) == 1)
#define IS_SCHEDULED(tid) (atomic_read(&srr_pt[tid].is_scheduled) != 0)
#define ACQUIRE(tid) atomic_add_unless(&srr_pt[tid].is_used, 1, 1)
#define INC_SCHEDULED(tid) (atomic_inc(&srr_pt[tid].is_scheduled))
#define DEC_SCHEDULED(tid) (atomic_dec(&srr_pt[tid].is_scheduled))
#define SET_UNSCHEDULED(tid) (atomic_set(&srr_pt[tid].is_scheduled, 0))

int SrrPt_DeleteAll(pid_t pid)
{
	int tid;

	for(tid = 0; tid < NR_TIMERS; tid++) {
		if(srr_pt[tid].owner == pid)
			SrrPt_Delete(pid, tid);
	}

	return 0;
}

int SrrPt_IsValid(pid_t pid, srr_timer_t tid)
{
	if(tid < 0 || tid >= NR_TIMERS)
		return 0;
	
	if(pid != srr_pt[tid].owner)
		return 0;
	
	return 1;
}

void SrrPt_Get(srr_timer_t tid, struct itimerspec* its)
{
	unsigned long remaining = srr_pt[tid].timer.expires;
#ifdef CONFIG_UTIME
	unsigned long usec = srr_pt[tid].timer.usec;
	unsigned long jiffies_u_now = jiffies_u;
	usec -= jiffies_u_now;
#endif

	/* Handle jiffie wrap-around */
	if(remaining > jiffies)
		remaining -= jiffies;
	else
		remaining = (0xffffffff - jiffies) + remaining + 1;

#ifdef CONFIG_UTIME
	if(usec < jiffies_u_now) {
		usec += USEC_PER_JIFFIES;
		remaining -= 1;
	}

	if(remaining<0){
		remaining=0;
		jiffies=0;
	}

	jiffies_and_usec_to_ts(remaining,jiffies,&its->it_value);
#else	
	jiffies_to_timespec (remaining, &its->it_value);
#endif
	its->it_interval = srr_pt[tid].timespec.it_interval;

	PDEBUG("SrrPt_Get(tid %d) its %ld,%ld %ld,%ld \n",
		tid,
		its->it_value.tv_sec, its->it_value.tv_nsec,
		its->it_interval.tv_sec, its->it_interval.tv_nsec
		);
}

void SrrPt_Set(srr_timer_t tid, struct itimerspec* its)
{
#ifdef CONFIG_UTIME
	unsigned long usec;
	unsigned long expires;

	ts_to_jiffies_and_usec(&its->it_value,&expires,&usec);

	expires+=jiffies;
	usec+=jiffies_u;

	if(usec>= USEC_PER_JIFFIES){
		expires += usec / USEC_PER_JIFFIES;
		usec %= USEC_PER_JIFFIES;
	}

	ts_to_jiffies_and_usec(&its->it_interval,&srr_pt[tid].interval,&srr_pt[tid].usec);

	srr_pt[tid].timespec = *its;
	srr_pt[tid].interval = timespec_to_jiffies(&its->it_interval);
#else
	unsigned long expires = jiffies + timespec_to_jiffies(&its->it_value);

	srr_pt[tid].interval = timespec_to_jiffies(&its->it_interval);
	srr_pt[tid].timespec = *its;
#endif

	/* If the conversion to jiffies gives us a zero interval despite a
       non-zero request, bump it up to one jiffy. */
	if (srr_pt[tid].interval == 0 &&
		(its->it_interval.tv_sec != 0 ||
		 its->it_interval.tv_nsec != 0))
		srr_pt[tid].interval = 1;

	PDEBUG ("SrrPt_Set(tid %d) its %ld,%ld %ld,%ld j %ld e %ld i %ld\n",
			tid,
			srr_pt[tid].timespec.it_value.tv_sec,
			srr_pt[tid].timespec.it_value.tv_nsec,
			srr_pt[tid].timespec.it_interval.tv_sec,
			srr_pt[tid].timespec.it_interval.tv_nsec,
			jiffies, expires, srr_pt[tid].interval
		);
	
	if(itimerspec_isset(its))
	{
		if (timer_pending(&srr_pt[tid].timer))
		{
			PDEBUG ("Timer %d armed while still pending. arm %d, fire %d\n",
					tid, srr_pt[tid].arm_count, srr_pt[tid].fire_count);
		}
	
		srr_pt[tid].arm_count++;
		
		/* If we are setting a timer that has already expired, which
		   is possible if srr_timespec_sub generated a negative number,
		   fire immediately.
		   CHANGED: AT (Sep 26/04) - the timer code in the kernel
		   deals with timers in the past by firing at the next cycle
		   so we do not have to. */
		/* Allow expires to be less than jiffies, which is a jiffie
		   wrap-around situation.  This happens once every 48.6 days.
		   AT (Jun 4/07)*/

		if (//expires < jiffies ||
			(long)srr_pt[tid].timespec.it_value.tv_nsec < 0)
		{
			PDEBUG("Timer %d fired in the past. arm %d, fire %d: %lu -> %lu\n",
				   tid, srr_pt[tid].arm_count, srr_pt[tid].fire_count,
				   expires, jiffies);
			
			expires = jiffies;
		}
		srr_pt[tid].expires = expires;

#ifdef CONFIG_UTIME
		del_timer(&srr_pt[tid].timer);
		srr_pt[tid].timer.expires=expires;
		srr_pt[tid].timer.usec=usec;
		add_timer(&srr_pt[tid].timer);
		PDEBUG ("Timer set within UTIME\n");
#else
		mod_timer(&srr_pt[tid].timer, expires);
		PDEBUG ("Timer set by mod_timer, now=%ld, expire=%ld\n",
				jiffies, expires);
#endif
		// AST April 30/08
		// If the timer is not pending after a call to mod_timer, it
		// has already fired.  That could only happen on a multi-core
		// system.  This check is not valid, and the Oops is not real.
		// Older kernels might refuse to add a timer if expires < jiffies
		// but that is also not valid since the timer could cross a
		// jiffie wrap-around boundary.
		//if (!timer_pending(&srr_pt[tid].timer))
		//	Oops ("Timer %d queued but not pending. arm %d, fire %d\n",
		//		  tid, srr_pt[tid].arm_count, srr_pt[tid].fire_count);
	}
	else
	{
		PDEBUG ("Stop timer during timer_set.  Zero time specified.\n");
		del_timer(&srr_pt[tid].timer);
	}
}

int SrrPt_Delete(pid_t pid, srr_timer_t tid)
{
	USES_TIMER_LOCK();
	PDEBUG ("Delete timer\n");
	
	if(!SrrPt_IsValid(pid, tid))
	{
		PDEBUG ("Deleting invalid timer\n");
		return -EINVAL;
	}

	/* stop the timer, null op if not running.
	 * Do it inside the spin lock so we do not delete the
	 * timer while the timer interrupt handler is running.
	 * It could race and be modify_timer and del_timer at
	 * the same time. */
	TIMER_LOCK();
	SET_UNUSED(tid);
	del_timer(&srr_pt[tid].timer);
	srr_pt[tid].owner = 0;
	TIMER_UNLOCK();

	return 0;
}

int SrrPt_Create( pid_t owner, struct sigevent* se )
{
	int tid;
	USES_TIMER_LOCK();
//	TDEBUG( "Creating timer for %d.  Signal is %d\n", owner, se->sigev_signo );
	PDEBUG( "Create timer\n" );
	
	TIMER_LOCK();
	for(tid = 0; tid < NR_TIMERS; tid++) {
		if(!IS_SCHEDULED(tid))
		{
			if( ACQUIRE( tid ) )
			{
				// Check to see if timer was scheduled while acquiring
				// It should never happen.
				if( !IS_SCHEDULED( tid ) )
				{
					break;
				}
				
				SET_UNUSED( tid );
			}
		}
	}

	if( tid == NR_TIMERS )
	{
		TIMER_UNLOCK();
		return -EAGAIN;
	}

	// memset(&srr_pt[tid], 0, sizeof(srr_pt[tid]));

	// set the things we know
	srr_pt[tid].owner = owner;
	srr_pt[tid].event = *se;
	srr_pt[tid].tid = tid;
	
	// srr_pt[tid].is_used is set by ACQUIRE

	// zero everything else
	atomic_set( &srr_pt[tid].qcount, 0 );
	SET_UNSCHEDULED( tid );
	srr_pt[tid].arm_count = 0;
	srr_pt[tid].fire_count = 0;
	srr_pt[tid].trigger_count = 0;
	srr_pt[tid].queued = 0;
	srr_pt[tid].queue_fail = 0;
	srr_pt[tid].bottom_halfs = 0;
	srr_pt[tid].clock = 0;
	srr_pt[tid].interval = 0;
	srr_pt[tid].usec = 0;
	srr_pt[tid].flags = 0;
	srr_pt[tid].expires = 0;		/* expiry time, in jiffies */
	memset( &srr_pt[tid].timespec, 0, sizeof(srr_pt[tid].timespec) );
	memset( &srr_pt[tid].qtask, 0, sizeof(srr_pt[tid].qtask) );
	
	// Is there any reason to protect more of this initialzation within
	// the lock?  A timer is "used" if it has an owner.  Just a question
	// of whether the timer can be destroyed before this function completes.
	// I don't think so.  The task would have to die for that to happen,
	// and that wouldn't happen until after this call completes, would it?
	// If it did happen, and a new timer were instantly created, the new
	// timer could get initialized with the old information.
	// Moved the unlock to the bottom anyway.

	init_timer( &srr_pt[tid].timer );
	srr_pt[tid].timer.data = tid;
	srr_pt[tid].timer.function = SrrPt_TimerExpiry;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	srr_pt[tid].qtask.next = NULL;
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	srr_pt[tid].qtask.sync = 0;
	srr_pt[tid].qtask.routine = SrrPt_BottomHalf;
	srr_pt[tid].qtask.data = &srr_pt[tid];
#else
#ifdef WORK_DATA_IS_VOID
	INIT_WORK( &srr_pt[tid].qtask, SrrPt_BottomHalf, &srr_pt[tid] );
#else
	INIT_WORK( &srr_pt[tid].qtask, SrrPt_BottomHalf );
#endif // WORK_DATA_IS_VOID
#endif
	TIMER_UNLOCK();
	return tid;
}

/*
 * (pseudo-)system calls
 */

int SrrModTimerCreate( struct SrrTm* map, struct SrrTask* task )
{
	SrrTimerCreateArgs* args = &task->args.tcreate;

	TDEBUG("SrrModTimerCreate(signo %d)\n", args->event.sigev_signo);

	return SrrPt_Create(task->pid, &args->event);
}

int SrrModTimerDelete( struct SrrTm* map, struct SrrTask* task )
{
	SrrTimerDeleteArgs* args = &task->args.tdelete;

	PDEBUG("SrrModTimerDelete(tid %d)\n", args->tid);

	return SrrPt_Delete(task->pid, args->tid);
}

int SrrModTimerSet(SrrTm* map, SrrTask* task)
{
	SrrTimerSetArgs* args = &task->args.tset;
	struct itimerspec value;

	if(cpy_u2k(&value, args->value, sizeof(value)))
		return -EFAULT;

	PDEBUG("SrrModTimerSet(tid %d, flags %#x, value %ld,%ld, interval %ld,%ld)\n",
		args->tid, args->flags,
		value.it_value.tv_sec, value.it_value.tv_nsec,
		value.it_interval.tv_sec, value.it_interval.tv_nsec
		);

	if(!SrrPt_IsValid(task->pid, args->tid))
		return -EINVAL;

	/* If the it_value component of the time is 0,0 we do not translate it,
	   but instead pass it through intact, where it will be disabled. */
	/* if absolute, change value to default (relative) */
	if (itimerspec_isset(&value) && (args->flags & TIMER_ABSTIME))
	{
		int		nonzero = (value.it_value.tv_sec || value.it_value.tv_nsec);
		struct timespec  orig;
		struct timespec* ts = &value.it_value;
		struct timeval tvcur;
		struct timespec tscur;

		orig = value.it_value;
		do_gettimeofday(&tvcur);

		srr_timeval_to_timespec(&tscur, &tvcur);

		srr_timespec_sub(ts, ts, &tscur); /* *ts = *ts - tscur */
		if (nonzero &&
			(ts->tv_sec == 0 && ts->tv_nsec == 0))
		{
			PDEBUG ("Timespec change to relative generated 0 result. Fixed.\n");
			ts->tv_nsec = 1;
		}

		PDEBUG("abs: val %ld,%ld -> rel %ld,%ld (cur %ld,%ld)\n",
			   orig.tv_sec, orig.tv_nsec,
			   ts->tv_sec, ts->tv_nsec,
			   tscur.tv_sec, tscur.tv_nsec);
	}

	/* get current value to return as old value */
	if(args->ovalue) {
		struct itimerspec ovalue;
		SrrPt_Get(args->tid, &ovalue);
		if(cpy_k2u(args->ovalue, &ovalue, sizeof(ovalue)))
			return -EFAULT;
	}

	/* set new value */
	SrrPt_Set(args->tid, &value);

	return 0;
}

int SrrModTimerGet( SrrTm* map, SrrTask* task )
{
	SrrTimerGetArgs* args = &task->args.tget;
	struct itimerspec value;
	

	PDEBUG("SrrModTimerGet(tid %d)\n", args->tid);

	if(!SrrPt_IsValid(task->pid, args->tid))
		return -EINVAL;

	SrrPt_Get(args->tid, &value);

	/* timer_get can be used to verify that timer is valid */
	if(args->value && cpy_k2u(args->value, &value, sizeof(value)))
		return -EFAULT;

	return 0;
}

int SrrModClockGetres(SrrTm* map, SrrTask* task)
{
	SrrClockGetresArgs* args = &task->args.cgetres;
	struct timespec value;

	PDEBUG("SrrModClockGetres(tid %d)\n", args->tid);

	if (args->tid != CLOCK_REALTIME)
		return -EINVAL;

	if (args->value)
	{

		value.tv_sec = 0;
		value.tv_nsec = 1000000000 / HZ;


		if(args->value && cpy_k2u(args->value, &value, sizeof(value)))
		   return -EFAULT;
	}

	return 0;
}

/* We are not checking this InvalidTid flag at the moment.  It could be used
   to generate some kind of debug message outside the interrupt handler if
   the timer expires with an invalid timer ID.  We are dealing with that
   anyway since the interrupt could occur while the timer is being
   destroyed. */
static atomic_t InvalidTid;

void SrrPt_TimerExpiry(unsigned long tid)
{
	unsigned long interval;
	int signo;

	SrrTimerCount++;

	/* can't use IsValid(), because the current pid has no relationship with
	   the timer owner pid */
	if (tid < 0 || tid >= NR_TIMERS
		|| !IS_USED(tid)
		|| !srr_pt[tid].timer.function
		)
	{
		// Make a note that the tid was invalid, but do not generate
		// an Oops inside the interrupt handler.
		atomic_set(&InvalidTid, 1);
	}
	else if (IS_USED(tid))
	{
		int		result=0;

		// Set the timer to be scheduled early so it will not be cleaned up
		// if deleted in another thread.
		INC_SCHEDULED(tid);

		// Then, if it was deleted while were were busy setting it
		// to be scheduled, do not actually do anything with the timer
		if (!IS_USED(tid))
		{
			DEC_SCHEDULED(tid);
		}
		else
		{
			atomic_set(&InvalidTid, 0);
			srr_pt[tid].fire_count++;
			interval = srr_pt[tid].interval;
			signo = srr_pt[tid].event.sigev_signo;
	
			atomic_inc (&srr_pt[tid].qcount);
			srr_pt[tid].queued++;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
			result = schedule_task (&srr_pt[tid].qtask);
#else
			result = schedule_work (&srr_pt[tid].qtask);
#endif
		
			// Set the timer to fire at the next interval
			if(interval)
			{

#ifdef CONFIG_UTIME
				unsigned long usec;
				unsigned long expires;

				expires = jiffies + srr_pt[tid].interval;

				usec = jiffies_u + srr_pt[tid].usec;

				if(usec>= USEC_PER_JIFFIES){
					expires += usec / USEC_PER_JIFFIES;
					usec %= USEC_PER_JIFFIES;
				}

				del_timer(&srr_pt[tid].timer);

				srr_pt[tid].timer.expires=expires;
				srr_pt[tid].timer.usec=usec;
				srr_pt[tid].expires = expires;
	
				add_timer(&srr_pt[tid].timer);
#else
				srr_pt[tid].expires = jiffies + interval;
				mod_timer(&srr_pt[tid].timer, srr_pt[tid].expires);
#endif
			}
			
			// The task was already queued.  Reduce the schedule count
			// again because we did not really schedule it.
			if (result == 0)
			{
				srr_pt[tid].queue_fail++;
				DEC_SCHEDULED(tid);
			}
		}
	}
}

/* We used to hold the spin lock in this code, but now we limit it
   to the timer lock so that we don't mess with the timers while the
   bottom half is running.  The module spin lock is only grabbed
   by the trigger function when it needs to alter the task list. */

#ifdef WORK_DATA_IS_VOID
void SrrPt_BottomHalf (void* data)
#else
void SrrPt_BottomHalf (struct work_struct *work)
#endif // WORK_DATA_IS_VOID
{
#ifdef WORK_DATA_IS_VOID
	SrrPt		*pt = (SrrPt*) data;
#else
	SrrPt		*pt = container_of (work, SrrPt, qtask);
#endif // WORK_DATA_IS_VOID
	int			err, valid=1;
	static int	entered = 0;
	USES_TIMER_LOCK();

	PDEBUG ("Timer fired at %ld\n", jiffies);
	
	if (in_interrupt())
	{
		Oops ("Timer bottom half called from interrupt\n");
		if (pt)
			DEC_SCHEDULED(pt->tid);
		return;
	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	if (irqs_disabled())
	{
		Oops ("IRQs disabled in timer bottom half handler\n");
	}
#endif
	
	if (entered)
	{
		// This is not really an Oops.  We use it for debugging.
		// Oops ("Timer bottom half reentered\n");
	}
	
	entered ++;

	if (!pt || !IS_USED(pt->tid))
	{
		valid = 0;
		
		if (!pt)
		{
			Oops ("Timer bottom half called on NULL timer\n");
		}
		else
		{
			/* On an SMP system it is possible for the pt pointer to
			 * refer to a timer that has already been deleted, so this
			 * is in fact a normal state.
			 */
			DEC_SCHEDULED(pt->tid);
		}
	}
	else if (pt->event.sigev_signo >= 0)
	{
		pt->bottom_halfs++;
		// Need to lock to avoid a race where this timer could
		// be deleted, setting the owner to zero
                TIMER_LOCK();
		if (IS_USED(pt->tid))
		{
			/* int err = */
			SrrTm_Kill(srr_taskmap, pt->owner,
					   pt->event.sigev_signo);
		}
		TIMER_UNLOCK();
		DEC_SCHEDULED(pt->tid);
	}
	else
	{
		/* Loop over the number of times that the timer has fired
		   since the last interrupt.  We do reads atomically because
		   there is a possibility that the timer might fire while we
		   are looping.  We might loop more than the number of times
		   that were pending when we started, and that would cause an
		   entry into this routine with 0 pending firings later. */

		pt->bottom_halfs++;
		while (atomic_read(&pt->qcount) > 0)
		{
			int proxy = -pt->event.sigev_signo;
			
			atomic_dec (&pt->qcount);
			pt->trigger_count++;
			err = SrrDoTrigger(srr_taskmap, proxy, 0);

			/* It is not an error to trigger a proxy that does not
			   exist because the proxy could be deleted before the
			   bottom half is scheduled. (AST)*/
			if (err != 0 && err != -ESRCH)
			{
				Oops ("Proxy %d timer returned error %d\n", proxy, err);
				atomic_set( &pt->qcount, 0); /* clear proxy trigger if proxy is deleted */
			}
		}
		DEC_SCHEDULED(pt->tid);
	}

	entered --;
}
