/*
 * srr_tt.c: implementation of the task structures
 *
 *  Linux Loadable Module Implementation of QNX Send/Receive/Reply
 *	message passing, and related function.
 *
 *	QNX is a trademark of QSSL, Ltd. <www.qnx.com>
 *
 * Copyright (C) 2000 Cogent Real-Time Systems Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software in a file called 'COPYING'; if not, write to:
 * 	Free Software Foundation, Inc.,
 * 	59 Temple Place, Suite 330,
 * 	Boston, MA 
 * 	02111-1307  USA
 *
 * $Id: srr_tt.c 340 2014-12-30 20:45:02Z mstormer $
 *
 * added define for NAME_SPECIFIC_LOCKS  and macros 
 *
 */

#include "module_versions.h"

#include "sysdep.h"

#include "srr_tt.h"
#include "srr_sys.h"
#include "srr_time.h"
#include "srr_util.h"

#include <linux/vmalloc.h>

// define maximum size for kalloc (in SrrBuf_AdjustSize())
// kalloc is limited to 128K (but that is not garanteed)
#define SRRBUF_MAX_KALLOC 65536  // 64Kb

#define CHKMAP(call) if(!map) { Oops(call " called with no map!\n"); return 0; }

#if defined( TASK_SPECIFIC_LOCK) 
DEFINE_MUTEX(proxyLock);
#endif

#if defined ( NAME_SPECIFIC_LOCKS)
DEFINE_MUTEX(nameLock);
#endif

extern spinlock_t rcv_q_lock;

/* If we use these definitions, do not use the ones in srr_sys.c.  They
   are here to protect the recv_q from damage when an interrupt occurs
   that causes a proxy to be fired.  They do not protect task->state,
   which is also necessary.  If you use the definitions in srr_sys.c
   along with these, you will get a deadlock on a multi-processor
   system. */

//#define Srr_UseSpinLock()	int _spinflags
//#define Srr_SpinLock()		spin_lock_irqsave(&rcv_q_lock, _spinflags)
//#define Srr_SpinUnlock()	spin_unlock_irqrestore(&rcv_q_lock, _spinflags)

#undef Srr_UseSpinLock
#undef Srr_SpinLock
#undef Srr_SpinUnlock
#define Srr_UseSpinLock()
#define Srr_SpinLock()
#define Srr_SpinUnlock()

Static const char* SrrTaskState2Str(SrrTask* task, int state)
{
	switch(state)
	{
		case RELEASED:		return "RELEASED";
	
  case READY:
			return (task && (task->flags & _PPF_POLLING) ? "POLL" : "READY");
		case RECV_BLOCKED:	return "RECV";
		case SEND_BLOCKED:	return "SEND";
		case REPLY_BLOCKED:	return "REPLY";
		case RECV_SPINLOCKED:	return "SPINR";
		case TRIGGER_SPINLOCKED:	return "SPINT";
		default:			return "<unknown>";
	}
}

Static SrrTask* SrrTask_new(pid_t pid)
{
	
  SrrTask* task = xxmalloc(sizeof(SrrTask));
  if(!task || task == ZERO_SIZE_PTR)
  {
		return 0;
  }
	
	memset(task, '\0', sizeof(SrrTask));

	task->pid	= pid;
	task->state	= READY;
	task->flags = 0;
	task->kern_task = find_task_by_pid(pid);

#ifdef TASK_SPECIFIC_LOCKS
#ifdef LOCK_VIA_SPINLOCKS
    spin_lock_init(&(task->mutex.lock));
#else
    mutex_init(&(task->mutex.mutex));
#endif
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)
	init_waitqueue_head(&task->poll_q);
	init_waitqueue_head(&task->wq);
#ifdef TASK_SPECIFIC_LOCKS  
  init_waitqueue_head(&task->mutex.wq);
#endif
#endif
	
	return task;
}

Static void SrrTask_delete(SrrTask* task, SrrTm *map, int recurse)
{
/* return mts */
#ifdef TASK_SPECIFIC_LOCKS
  SrrMutex *mutexPtr = &task->mutex;
  if(!recurse)
    cleanDeleteQueue(map);
  
  if(atomic_read(&mutexPtr->users)){
    DQDEBUG("Task_Delete(%d): Trying to free mutex with %d users!\n", task->pid, atomic_read(&mutexPtr->users)); 
    addToDeleteQueue(task, map);
    return;
  }
  else {
#ifndef LOCK_VIA_SPINLOCKS
	mutex_destroy(&mutexPtr->mutex);
#endif
#endif
//	printk(KERN_DEBUG "Task_delete: Deleting pid %d.\n", deleted);
  /* Free xfer buffer - 2000/07/24 Alessandro Sala */
	if (task->buf.data){
          // use either kfree or vfree depending on buffer size ...
          if (task->buf.maxsize < SRRBUF_MAX_KALLOC)
		kfree(task->buf.data);
          else
		vfree(task->buf.data);
        }

	/* Remove all IRQ handlers if they have not already been removed
	   by a normal task exit (this would happen if we unloaded the
	   module with a non-zero usage count) */
#ifndef NO_IRQ
	SrrIrq_DetachAll (task);
#endif
	memset(task, 0, sizeof(*task));
	xxfree(task);
#ifdef TASK_SPECIFIC_LOCKS
  }
#endif
  
}

/*
 * SrrBuf
 */

/* 2000/04/01 - Alessandro Sala
 * Adjusts the length of a buffer, allocating
 * a new one if the current length is less than "len"
 * and preserving the old content
 *
 * Returns 1 on success, 0 on failure.
 */
int SrrBuf_AdjustSize(SrrBuf* buf,  size_t len)
{
  
	 NDEBUG("SrrBuf_AdjustSize()\n");

	/* ensure buffer length is >= to len */
	if(buf && len > buf->maxsize) 
	{
		void *newdata;

		// use kmalloc/kfree for small messages and vmalloc/vfree for huge ones
		if( len < SRRBUF_MAX_KALLOC )
		{
			if( !( newdata = kmalloc( len, GFP_ATOMIC ) ) )
				return 0;
		}
		else
		{
			if( !( newdata = vmalloc( len ) ) ) 
				return 0;
		}

		if( buf->data ) 
		{
			if( memcpy( newdata, buf->data, buf->datalen ) == ZERO_SIZE_PTR )
			{
				if( buf->maxsize < SRRBUF_MAX_KALLOC )
					kfree( newdata );
				else
					vfree( newdata );
                          
				return 0;
			}
			// use kfree/vfree depending on old buffer size
			if( buf->maxsize < SRRBUF_MAX_KALLOC )
				kfree( buf->data ); 
			else
				vfree( buf->data ); 
        }

		buf->data = newdata;
		buf->maxsize = len;
	}
  
	return 1;
}

void SrrBuf_Reset (SrrBuf* buf)
{
	buf->datalen = 0;
}

/* 2000/03/26 - Alessandro Sala
 * Reads form user space into a kernel buffer, starting at "offset"
 * and stopping at the *physical* end of the buffer (i.e. does not
 * grow the buffer). Updates the logical size of the buffer if the
 * new size is greater than old one.
 *
 * Returns the number of bytes read.
 */
int SrrBuf_OffsetRead(SrrBuf* buf, size_t offset, char* from, size_t size)
{
  
	if( !buf ) 
	{
		Oops( "can't offset read into a null buffer!\n" );
    	return -ENOMEM;
	}

	NDEBUG("SrrBuf_OffsetRead()\n");

	/* ensure buffer len is >= to size */
	if(size + offset > buf->maxsize)
	{
		if (!SrrBuf_AdjustSize(buf, size + offset))
		{
			PDEBUG("SrrBuf_OffsetRead: Could not adjust buffer size to %u\n",
				 (size + offset));
      
			return -ENOMEM;
		}
	}

	if (size > 0)
	{
		cpy_u2k(buf->data + offset, from, size);
		PDEBUG("SrrBuf_OffsetRead: Copy %d bytes at offset %d to buffer: %s\n",
			   size, offset, ((char*)buf->data) + offset);

/**
 *  ME: This next section will cause a kernel panic with
 * *Timeout functions enabled.  Why, I have no idea.
 **/
#ifdef NOT_EVER_DEFINED_EVER_IF_YOU_KNOW_WHATS_GOOD_FOR_YOU
#ifdef SRR_DEBUG
		{
			int		i;
			for (i=0; i<size; i++)
				if (from[0] != from[i])
					break;
			PDEBUG ("Input buffer is same to %d bytes\n", i);
			for (i=0; i<size; i++)
				if (((char*)buf->data)[offset+i] != from[i])
					break;
			PDEBUG ("Output buffer is same to %d bytes\n", i);
		}
#endif
#endif
	}
	if (buf->datalen < offset + size)
		buf->datalen = offset + size;
  
	return (size);
}

int SrrBuf_Read(SrrBuf* buf, char* from, size_t size)
{
	buf->datalen = 0;
	return (SrrBuf_OffsetRead (buf, 0, from, size));
}

/* 2000/03/27 - Alessandro Sala
 * Writes into user space from a kernel buffer, starting at "offset"
 * and stopping at the *physical* end of the buffer.
 *
 * Returns the number of bytes written.
 */
 
 
int SrrBuf_OffsetWrite(SrrBuf* buf, size_t offset, char* to, size_t size)
{
  
	int k2u;
	
	if(!buf) {
		Oops("can't offset write from a null buffer!\n");
    
		return 0;
	}
	NDEBUG("SrrBuf_OffsetWrite()\n");

	/* attempt to read past the end of buffer */
	if (offset >= buf->datalen)
	{
#ifdef SRR_DEBUG
		if ( buf->datalen > 0 )
			PDEBUG("SrrBuf_OffsetWrite: read past end of input buffer\n");
#endif
		return 0;
	}

	/* truncate read at input buffer end.  This is normal operation
	 * since the TO buffer will be larger than the FROM buffer in most
	 * cases.  So, the maximum read is buf->datalen - offset;
	 */
	if(size > buf->datalen - offset)
	{
		size = buf->datalen - offset;
	}
#if 0
	PDEBUG ("Copy kernel to user: 0x%lx %d,%d,%d,%d for %u bytes, offset %u\n",
			(size_t)to,
			((char*)buf->data)[offset+0],
			((char*)buf->data)[offset+1],
			((char*)buf->data)[offset+2],
			((char*)buf->data)[offset+3],
			size, offset);
#endif
	k2u = cpy_k2u(to, buf->data + offset, size);
	if (k2u < 0)
	{
		PDEBUG ("Copy kernel to user: 0x%x %d,%d,%d,%d for %u bytes, offset %u\n",
			  (size_t)to,
			  ((char*)buf->data)[offset+0],
			  ((char*)buf->data)[offset+1],
			  ((char*)buf->data)[offset+2],
			  ((char*)buf->data)[offset+3],
			  size, offset);
        Oops("Error k2u during SrrBuf_OffsetWrite");
	}

//ifdef  SRRDEBUG
#if  0 
	{
		int		i, j=0;
		PDEBUG ("Attempting buffer access.\n");

		for (i=j=0; i<size; i++)
			j += ((char*)buf->data)[i+offset];
		PDEBUG ("Accessed kernel buffer OK: j=%d.\n", j);

		for (i=j=0; i<size; i++)
			j += to[i];
		PDEBUG ("Accessed output buffer OK: j=%d.\n", j);
		
		for (i=0; i<size; i++)
			if (((char*)buf->data)[0+offset] != ((char*)buf->data)[i+offset])
				break;
		PDEBUG ("Kernel buffer is same to %d bytes\n", i);

		for (i=0; i<size; i++)
			if (((char*)buf->data)[offset+i] != to[i])
				break;
		PDEBUG ("Output buffer is same to %d bytes\n", i);
	}
#endif
  
	return size;
}

int SrrBuf_Write(SrrBuf* buf, char* to, size_t size)
{
	return (SrrBuf_OffsetWrite (buf, 0, to, size));
}

/*
 * Read and write locally, rather than between kernel and user space
 */

int SrrBuf_LocalRead(SrrBuf* buf, char* from, size_t size)
{
  
	if(!buf) {
		Oops("can't read into a null buffer!\n");
		return -EINVAL;
	}

	NDEBUG("SrrBuf_LocalRead()\n");

	buf->datalen = 0;

	/* ensure buffer len is >= to size */
	if(size > buf->maxsize) 
	{
		if (!SrrBuf_AdjustSize(buf, size))
		{
			return -ENOMEM;
    	}
	}

	memcpy(buf->data, from, size);
	buf->datalen = size;
  
	return size;
}

int SrrBuf_LocalWrite(SrrBuf* buf, char* to, size_t size)
{	
  
	if(!buf) {
		Oops("can't write from a null buffer!\n");
    
		return (-EINVAL);
	}

	NDEBUG("SrrBuf_LocalWrite()\n");

	if(size > buf->datalen)
		size = buf->datalen;
	
	memcpy(to, buf->data, size);
  
	return (size);
}

/*** Rq ***/

SrrTask** SrrRq_First (SrrTask* server)
{
	SrrTask		**first;
	//	Srr_UseSpinLock();

	//	Srr_SpinLock();
	first = &server->recv_q;
	//	Srr_SpinUnlock();
	return (first);
}

int SrrRq_NonEmpty (SrrTask* server)
{
	int			result;
	//	Srr_UseSpinLock();

	//	Srr_SpinLock();
	result = server->recv_q == NULL ? 0 : 1;
	//	Srr_SpinUnlock();
	return (result);
}

SrrTask** SrrRq_Find(SrrTask* server, pid_t pid)
{
	SrrTask**	taskp;
	Srr_UseSpinLock();

//	Srr_SpinLock();
	taskp = &server->recv_q;
	while(*taskp != 0 && (*taskp)->pid != pid)
		taskp = &(*taskp)->recv_l;
	if (pid == 0 && *taskp != NULL)
		Oops ("%s: Looking for queue end, found %d\n",
			  __FUNCTION__, (*taskp)->pid);
//	Srr_SpinUnlock();
	return taskp;
}

extern struct SrrTm* srr_taskmap;

EXPORT_SYMBOL( srr_taskmap );  /* added to expose interface to other kernel modules */

#ifdef SRR_DEBUG
/* This is a debugging function that looks for proxies that have a non
   zero count, but which are not send-blocked on their target.  This
   represents a bug in the receive queueing code if it happens. */
int SrrRq_CheckLinks (char* where)
{
	int			i, invalid=0;
	SrrTask* 	t;
	SrrTask**	pt;
	static int	reported = 0;

	//	if (!reported)
	{
		for(i = 0; i < NR_TASKS; ++i)
		{
			t = srr_taskmap->tasks[i];
		
			if(t && IS_PROXY(t))
			{
				if (t->count.counter != 0 &&
					(!(pt = SrrRq_Find(t->target, t->pid)) || !*pt))
				{
					Oops ("%s: proxy %d has count %d and not queued\n",
						  where, t->pid, t->count.counter);
					reported = 1;
					invalid = 1;
				}
			}
		}
	}
	return (invalid);
}
#endif

SrrTask* SrrRq_UnlinkNdebug(SrrTask** taskp)
{
	SrrTask*	task;
	Srr_UseSpinLock();

	Srr_SpinLock();
	task = *taskp;
	*taskp = task->recv_l;
	task->recv_l = 0;
	Srr_SpinUnlock();
	
	return task;
}

SrrTask* SrrRq_Unlink(SrrTask** taskp)
{
	SrrTask*	task;

#ifdef SRR_DEBUG
	SrrRq_CheckLinks ("Unlink Start");
#endif

	task = SrrRq_UnlinkNdebug(taskp);

#ifdef SRR_DEBUG
	SrrRq_CheckLinks ("Unlink End");
#endif
	
	return task;
}

extern int SrrTimerCount;

/* queue task onto server's receive queue */
void SrrRq_Queue (SrrTask* server, SrrTask* task)
{
#ifdef SRR_DEBUG
	SrrRq_CheckLinks ("Queue Start");
#endif
	SrrRq_QueueNdebug (server, task);
#ifdef SRR_DEBUG
	SrrRq_CheckLinks ("Queue End");
#endif
}

void SrrRq_QueueNdebug (SrrTask* server, SrrTask* task)
{
	SrrTask		**taskp;
	Srr_UseSpinLock();
#ifndef TASK_SPECIFIC_LOCKS
	Srr_SpinLock();
#endif

	taskp = &server->recv_q;
	while(*taskp != 0 && (*taskp)->pid != 0)
	{
		if (task == *taskp || task->pid == (*taskp)->pid)
			Oops ("Try to add a task %d already on receive queue of %d\n",
				  task->pid, server->pid);
		taskp = &(*taskp)->recv_l;
	}
	if (*taskp != NULL)
		Oops ("%s: Looking for queue end, found %d\n",
			  __FUNCTION__, (*taskp)->pid);
	*taskp = task;
#ifndef TASK_SPECIFIC_LOCKS
	Srr_SpinUnlock();
#endif
}

/*** Tm ***/

SrrTm* SrrTm_new()
{
	SrrTm* map;
	map = xxmalloc(sizeof(SrrTm));

	if(!map || map == ZERO_SIZE_PTR){
		return 0;
  	}
	
	memset(map, '\0', sizeof(SrrTm));

	map->tasks = xxmalloc(sizeof(SrrTask*) * NR_TASKS);

	if(!map->tasks || map->tasks == ZERO_SIZE_PTR) {
		xxfree(map);
		return 0;
	}

	memset(map->tasks, '\0', sizeof(SrrTask*) * NR_TASKS);

#ifdef TASK_SPECIFIC_LOCKS
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)
	init_waitqueue_head(&map->Map_wq);
#endif

  	map->deleteQueue = xxmalloc(sizeof(struct SrrDeleteQueue));

  	if(!map->deleteQueue || map->deleteQueue == ZERO_SIZE_PTR)
  	{
    	xxfree(map->tasks);
    	xxfree(map);
    	return 0;
  	}

  	memset(map->deleteQueue, '\0', sizeof(struct SrrDeleteQueue));
  	map->deleteQueue->head = NULL;
  	map->deleteQueue->tail = NULL;
  	mutex_init(&map->deleteQueue->lock);
#endif
  	return map;
}

void SrrTm_delete(SrrTm* map)
{
	int i;
	
	AddMapWriter();
  	TTP();
  /* shouldn't be any tasks in the map if the map is being deleted */
	for(i = 0; i < NR_TASKS; ++i) {
		if(map->tasks[i]) {
			Oops("task map still had pid %d!\n", map->tasks[i]->pid);
			/* Better see if any timers are running.  If so, stop them
			   before we have a disaster. */
//			SrrStopTimedProxy (map->tasks[i]);
			SrrTask_delete(map->tasks[i], map, 0);
			map->tasks[i] = 0;
		}
	}
  	TTP();
#ifdef TASK_SPECIFIC_LOCKS
  	while(cleanDeleteQueue(map))
  	{
    	msleep(1);
  	}
  	xxfree(map->deleteQueue);
#endif
  	RemoveMapWriter();
  	xxfree(map->tasks);
	xxfree(map);
}

Static SrrTask* SrrTm_Get(SrrTm* map, pid_t pid)
{
  
	/* HASH ACCESS */
	int i = pid;
  	int accessCount = 0;

	/* LINEAR ACCESS
    int i;
	 */
	/* LINEAR ACCESS METHOD
	for(i = 0; i < NR_TASKS; ++i) {
		if(map->tasks[i] && map->tasks[i]->pid == pid)
			return map->tasks[i];
	}
	*/
#ifndef DONT_LOCK_GET
  	AddMapReader();
#endif
  	while(1)
	{
		if(map->tasks[i % NR_TASKS])
		{
			if(map->tasks[i % NR_TASKS]->pid == pid)
			{
#ifndef DONT_LOCK_GET
  				RemoveMapReader();
#endif
        		return map->tasks[i % NR_TASKS];
			}
		}
		
		if(accessCount == NR_TASKS)
		{
#ifndef DONT_LOCK_GET
  			RemoveMapReader();
#endif
			return 0;
		}
		accessCount++;
		i++;
	}
#ifndef DONT_LOCK_GET
  	RemoveMapReader();
#endif
  
	return 0;
}

static pid_t	lowest	= 0x00010000;
static pid_t	highest	= 0x00ff0000;
static pid_t	last	= 0x00010000;

SrrTask* SrrTm_ProxyAlloc(SrrTm* map)
{
	pid_t     try   = last;
	SrrTask*  proxy = 0;

	ProxyAllocLock(); 
	AddMapWriter();

	while(1) 
	{
		try++;
		
		if(try == highest)
			try = lowest;

		if(try == last)
			break;
		
		/* Yes, SrrTm_Get is called within SrrTm_Alloc, but if we rely
		   on it there, then the way this code is written we attempt to
		   allocate the proxy 16 million times, complete with Oops()
		   calls, which hangs Linux very effectively. */
		if (!SrrTm_GetLocked (map, try))
		{
			proxy = SrrTm_Alloc(map, try, 1);
			break;
		}
	}
	last = try;
	
  	RemoveMapWriter();
  	ProxyAllocUnlock();
	return proxy;
}

SrrTask* SrrTm_Alloc(SrrTm* map, pid_t pid, int proxy)
{
  
	SrrTask* task;
	/* HASH ACCESS */
	int i = pid;
	int accessCount = 0;

	/* LINEAR ACCESS
	int i;
	 */
#ifdef TASK_SPECIFIC_LOCKS
	if(proxy)
	{
		RemoveMapWriter();
	}
#endif

	if(SrrTm_Get(map, pid))
	{
		TCDEBUG("Attempting to alloc pre-existing process %d.  Returning 0\n", pid);
    	return 0;
  	}
	
  	AddMapWriter();
	
  	task = SrrTask_new(pid);
	if(!task) 
	{
    	if(!proxy)
     		RemoveMapWriter();
	  	TCDEBUG("Task_new did not allocate new task for %d\n", pid);	
    	return 0;
	}
  	TCDEBUG("Task %d created!\n", pid);	
/*  LINEAR ACCESS METHOD
	for(i = 0; i < NR_TASKS; ++i) {
		if(!map->tasks[i]) {
			map->tasks[i] = task;
			break;
		}
	}
*/

/* HASH ACCESS METHOD
 * Uses linear probing */
	while(map->tasks[i % NR_TASKS])
	{
		if(accessCount == NR_TASKS)
		{
			Oops("srr: task table full, probably too many proxies allocated!\n");
			SrrTask_delete(task, map, 0);
			task = 0;

#ifdef TASK_SPECIFIC_LOCKS
      		if(!proxy)
      		{
        		RemoveMapWriter();
      		}
#endif
			return task;
		}
		i++;
		accessCount++;
	}
  	HASHDEBUG("Adding pid %d at hash %d\n", pid, i % NR_TASKS);
	map->tasks[i % NR_TASKS] = task;
/* END HASH ACCESS */
/*
	if(i == NR_TASKS) {
		* this is impossible, because our map is the same size as
		   the kernel's.
		   Wrong: We use the same task table for both proxies and
		   tasks, so we can fill it just by allocating lots of
		   proxies. (AST)

		Oops("srr: task table full, probably too many proxies allocated!\n");
		SrrTask_delete(task);
		task = 0;
	}
*/
  	TCDEBUG("Task %d added!\n", pid);	
#ifdef TASK_SPECIFIC_LOCKS
  	if(!proxy)
  	{
    	RemoveMapWriter();
  	}
#endif
  
	return task;
}

int SrrTm_Free(SrrTm* map, pid_t pid)
{
  
	SrrTask* task = 0;
#ifdef TASK_SPECIFIC_LOCKS
	SrrMutex *mutex;
#endif
	int		i = pid;
	int accessCount = 0;
	int		tskindx;
	Srr_UseSpinLock();

	KDEBUG("SrrTm_Free(%d)\n", pid);
#ifndef TASK_SPECIFIC_LOCKS
	Srr_SpinLock();
#else
  	AddMapWriter();
#endif
	/* find our task structure */
	/*
	for(i = 0; i < NR_TASKS; ++i) {
		if(map->tasks[i] && map->tasks[i]->pid == pid)
			break;
	}
	*/
	while(!map->tasks[i % NR_TASKS] || map->tasks[i % NR_TASKS]->pid != pid)
	{
		if(accessCount == NR_TASKS)
		{
#ifndef TASK_SPECIFIC_LOCKS
			Srr_SpinUnlock();
#else
      		RemoveMapWriter(); 
#endif
			return 0;
		}
		i++;
		accessCount++;
	}
	/*
	if(i == NR_TASKS)
	{
		Srr_SpinUnlock();
		return 0;
	}
*/
	HASHDEBUG("found task %d at hash %d\n", pid, i % NR_TASKS);
	task = map->tasks[i % NR_TASKS];
#ifdef TASK_SPECIFIC_LOCKS
	mutex = &task->mutex;
	atomic_inc(&mutex->users);
	Srr_TaskLock();
#endif
	tskindx = i;

	task->flags |= _PPF_TERMING;
	
	/* Remove interrupt handlers first, outside the spin lock, because
	   a well-timed interrupt might ask for the spin lock as well.  But
	   that is the whole point of having the spin lock.  */
#ifndef NO_IRQ
//	Srr_SpinUnlock();
	SrrIrq_DetachAll (task);
//	Srr_SpinLock();
#endif

	/* delete any timers.  This has to come before proxies because
	   otherwise the timer could fire between the time we delete the
	   proxy and the time we delete the timers. */
	PDEBUG("free timers (%d)\n", task->pid);
	SrrPt_DeleteAll(task->pid);

	/* detach our proxies */
	for(i = 0; i < NR_TASKS; ++i) 
	{
		SrrTask* t = map->tasks[i];

		if(t && t->target && t->target->pid == pid)
		{
#ifdef TASK_SPECIFIC_LOCKS
      		RemoveMapWriter();
      		Srr_TaskUnlock();
#else
			Srr_SpinUnlock();
#endif
			SrrTm_Free(map, t->pid);
#ifdef TASK_SPECIFIC_LOCKS
			AddMapWriter();
			Srr_TaskLock();
#else
			Srr_SpinLock();
#endif
		}
	}

	/* if we're on a tasks recv_q, get off it */
	if(task->state == SEND_BLOCKED) 
	{
		SrrTask* server = SrrTm_GetLocked(map, task->args.pid);
		if(!server)
      	{
			Oops("proxy? %d recv blocked on non-existent %d!\n",
				IS_PROXY(task), task->args.pid);
      	}
		else
		{
			SrrTask** taskp = SrrRq_Find(server, task->pid);
			if(!taskp || *taskp != task)
         	{
				Oops("not on server's queue!\n");
         	}
			else
			{
				if (IS_PROXY (task))
					atomic_set (&task->count, 0);
				SrrRq_Unlink(taskp);
			}
			task->state = READY;
		}
	}

	/*
	 * Wake any tasks blocked on this one.
	 */

	PDEBUG("release dependents (%d)\n", pid);

	for(i = 0; i < NR_TASKS; ++i) 
	{
		SrrTask* t = map->tasks[i];

		if(!t || t == task)
			continue;
		
		if(t->state == SEND_BLOCKED && t->args.pid == task->pid) 
		{
			/* t would like to send us a mesg */
			SrrTask** taskp = 0;
			
			KDEBUG("release task %d from state %16s\n",
				t->pid, SrrTaskState2Str(t, t->state));

			/* How is it possible to have a NULL *taskp here? */
			taskp = SrrRq_Find(task, t->pid);
			if (taskp && *taskp)
			{
				if (IS_PROXY ((*taskp)))
					atomic_set (&(*taskp)->count, 0);
				SrrRq_Unlink(taskp);
			}
			else
				Oops ("B: task %d is send blocked on %d, but %d not queued\n",
					  t->pid, t->args.pid, t->pid);
			
			SrrTaskWake(t, RELEASED, __FUNCTION__);
		} else if(t->state == RECV_BLOCKED && t->args.pid == task->pid) {
			/* t is waiting to recv from us, tant pis */
			
			KDEBUG("release task %d from state %16s\n",
				t->pid, SrrTaskState2Str(t, t->state));
			
			SrrTaskWake(t, RELEASED, __FUNCTION__);
		} else if(t->state == REPLY_BLOCKED && t->args.pid == task->pid) {
			/* t is hoping to recv a reply from us, won't happen */
			
			KDEBUG("release task %d from state %16s\n",
				t->pid, SrrTaskState2Str(t, t->state));
			
			SrrTaskWake(t, RELEASED, __FUNCTION__);
		}
	}

	/* detach any names */
	PDEBUG("free names (%d)\n", task->pid);
	SrrNm_Free(task->pid);

	/* Finally, check to see if any informed tasks need to know */
	if (!IS_PROXY(task)) 
	{
		for(i = 0; i < NR_TASKS; ++i) 
		{
			SrrTask* t = map->tasks[i];
			
			if (t && t != task && (t->flags & _PPF_INFORM)) 
			{
				/* Somebody needs to know we are dead */

				char	rmsg[8]; // why an rmsg?
				char	smsg[8];
				int		tpid, result,cnt=0;
			
				// the death message happens to be all zeros... 
				memset (smsg, 0, sizeof(smsg));

				tpid = t->pid;
				task->args.send.pid = tpid;
				task->args.send.ssize = sizeof (smsg);
				task->args.send.rsize = sizeof (rmsg);
				task->args.send.smsg = smsg;
				task->args.send.rmsg = rmsg;

				KDEBUG("tell %d we died (%d)\n", tpid, task->pid);
#ifdef TASK_SPECIFIC_LOCKS
				Srr_TaskUnlock();
#else
				Srr_SpinUnlock();
#endif
				/* If the send fails due to a signal, do it again.  We
				   have no way to repeat the message in user code since
				   this message is initiated by the SRR kernel module */
				do
				{
					/* During a free, we might in fact be freed by
					   another task after the real task has died.  We
					   re-find the kernel task pointer, which might be
					   NULL if the task is already gone.  We must do
					   this immediately before the clear signals in
					   this case since SrrDoSend can call schedule()
					   indirectly, possibly causing the task to vanish
					   while this loop runs. */
					task->kern_task = find_task_by_pid(pid);
					
					/* Clear all pending signals to the dying task.
					   We no longer assume that the current task is
					   the dying task, since this is patently false in
					   Linux.  If we do not clear pending signals we
					   end up in an infinite loop trying to re-send
					   the death message. */
					SrrClearPendingSignals (task->kern_task);
					RemoveMapWriter();	
					result = SrrDoSend (map, task, SrrBuf_LocalRead,
										SrrBuf_LocalWrite);
					AddMapWriter();
					if( result == -EINTR )
					{
						cnt++;
						/* pause giving task(s) a chance to die */
						schedule_timeout( cnt );

						if( (cnt == 1) || (cnt > 10) ) /* report 1st and excessive */
							Oops ("death message interrupted: %d %d \n", result, cnt );
                                           
						if( cnt > 10 ) 
							break;
					}
				} while (result == -EINTR);
#ifdef TASK_SPECIFIC_LOCKS
				Srr_TaskLock();
#else
				Srr_SpinLock();
#endif
				KDEBUG("  %d tell %s (%d)\n", tpid,
					   result == 0 ? "succeeded" : "failed",
					   task->pid);
			}
		}
	}

	map->tasks[tskindx % NR_TASKS] = 0;

#ifdef TASK_SPECIFIC_LOCKS
	Srr_TaskUnlock();
	atomic_dec(&mutex->users);
#endif
	/* free the data structures */
	KDEBUG("delete the task (0x%x, %d)\n", (unsigned int)task, pid);
  	TCDEBUG("Task %d being deleted...\n", task->pid);
	SrrTask_delete(task, map, 0);
#ifndef TASK_SPECIFIC_LOCKS
	Srr_SpinUnlock();
#else
    RemoveMapWriter();
#endif
	
	return 1;
}

int SrrTm_Invariant(SrrTm* map)
{
	return 0;
}

int SrrTm_Kill(SrrTm* map, pid_t pid, int signo)
{
  	int	i;
	int err = 0;
  	int accessCount = 0;
	
	if(pid == 0)
	{
		err = -ESRCH;
		for(i = 0; i < NR_TASKS; ++i) 
		{
			SrrTask* t = map->tasks[i];

			if(!t)
				continue;

			if (!(t->flags & _PPF_TERMING) && !(IS_PROXY(t)) && t->kern_task) 
			{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25))
				err = send_sig(signo, t->kern_task, 0);
#else
				err = kill_proc(t->pid, signo, 1);
#endif
				if(err)
					break;
			}
			else
			{
				KDEBUG ("Kill signal %d on terminating task %d\n",
						signo, t->pid);
			}
		}
	}
	else
	{
		i = pid;
		while(!map->tasks[i % NR_TASKS] || map->tasks[i % NR_TASKS]->pid != pid)
		{
			if(accessCount == NR_TASKS)
			{
				Oops("srr: Trying to kill non-existent process!\n");
				err = -ESRCH;
				break;
			}
			i++;
			accessCount++;
		}
    	HASHDEBUG("Found task %d at hash %d\n", pid, i % NR_TASKS);
		if(!err)
		{
			SrrTask *t = map->tasks[i % NR_TASKS];
      		if(IS_PROXY(t))
      		{
        		Oops("Attempting to kill proxy %d!\n", pid);
        		return -EINVAL;
      		}
			else if(!(t->flags & _PPF_TERMING) && t->kern_task)
			{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25))
        		err = send_sig(signo, t->kern_task, 0);
#else
				err = kill_proc(t->pid, signo, 1);
#endif
			}
			else
			{
				KDEBUG ("Kill signal %d on terminating task %d\n",
							signo, t->pid);
			}
		}
	}

	return err;
}

/*
 * SrrNm(name map): underlying attach/detach/locate implementation
 */

struct SrrNameIdMap srr_NameIdMap[SRR_MAX_NAMES] = { { NULL, NULL } };

int SrrNm_Attach(SrrTask* task, char* name)
{
	int	name_id, slot=-1;
	
   //NameLock();
   NameLock();

	for(name_id = 0; name_id < SRR_MAX_NAMES; ++name_id) {
		if(slot == -1 && !srr_NameIdMap[name_id].task)
			slot = name_id;
		if(srr_NameIdMap[name_id].name &&
		   strcmp(name, srr_NameIdMap[name_id].name) == 0)
		{
			NameUnlock();
			return (-EEXIST);
		}
	}
	if(slot == -1)
   {
		NameUnlock();
		return -EAGAIN;
   }
	srr_NameIdMap[slot].name = name;
	srr_NameIdMap[slot].task = task;

   NameUnlock();
	return slot;
}

int SrrNm_Locate(SrrTm* map, const char* name)
{
	int			name_id;
	SrrTask*	task;

	if(!name)
		return -EINVAL;
	
	for(name_id = 0; name_id < SRR_MAX_NAMES; ++name_id) {
		if(srr_NameIdMap[name_id].name &&
				strcmp(name, srr_NameIdMap[name_id].name) == 0)
			break;
	}
	if(name_id == SRR_MAX_NAMES)
		return -ESRCH;

	task = SrrTm_Get (map, srr_NameIdMap[name_id].task->pid);
	if (!task || (task->flags & _PPF_TERMING))
	{
		KDEBUG ("Name locate '%s' on terminating task %d\n",
				name, task ? task->pid : 0);
		return -ESRCH;
	}
	
	return srr_NameIdMap[name_id].task->pid;
}

int SrrNm_Detach(SrrTask* task, int name_id)
{
	SrrTask* holder = srr_NameIdMap[name_id].task;

		/* name must belong to task */
	if(!holder || holder != task ||
		name_id < 0 || name_id >= SRR_MAX_NAMES ||
			srr_NameIdMap[name_id].name == 0)
				return -EINVAL;

   NameLock();
	xxfree(srr_NameIdMap[name_id].name);
	srr_NameIdMap[name_id].name = 0;
	srr_NameIdMap[name_id].task = 0;
   NameUnlock();
	return 0;
}

void SrrNm_Free(pid_t pid)
{
	int name_id;

	NDEBUG("SrrNm_Free(%d)\n", pid);
	
	for(name_id = 0; name_id < SRR_MAX_NAMES; ++name_id) {
		if(srr_NameIdMap[name_id].task &&
				srr_NameIdMap[name_id].task->pid == pid) {
			
			NameLock();
				xxfree(srr_NameIdMap[name_id].name);
				srr_NameIdMap[name_id].name = 0;
				srr_NameIdMap[name_id].task = 0;
			NameUnlock();
		}
	}
	
}

#ifdef TASK_SPECIFIC_LOCKS
void addToDeleteQueue(struct SrrTask *task, SrrTm *map)
{
  
  struct DeleteNode *new = xxmalloc(sizeof(struct DeleteNode));

   DQDEBUG("Adding %d to delete queue\n", task->pid);

  memset(new, '\0', sizeof(struct DeleteNode));
  new->task = task;
  new->next = NULL;

  mutex_lock(&map->deleteQueue->lock); 

  if(map->deleteQueue->head == NULL)
    map->deleteQueue->head = new;
  
  if(map->deleteQueue->tail != NULL)
    map->deleteQueue->tail->next = new;
  
  new->parent = map->deleteQueue->tail;
  map->deleteQueue->tail = new;
  DQDEBUG("Add Queue Print:\n");
  PRINT_QUEUE_STATE();
  mutex_unlock(&map->deleteQueue->lock);
}

int cleanDeleteQueue(SrrTm *map)
{
  	struct DeleteNode *cur = NULL;
  	struct DeleteNode *next = NULL;
  	int missed = 0;

  	mutex_lock(&map->deleteQueue->lock);
  	DQDEBUG("Starting queue clearing...\n");
  	PRINT_QUEUE_STATE();
  	cur = map->deleteQueue->head;
  
  	while(cur != NULL)
  	{
    	next = cur->next;
    	if(cur->task && !atomic_read(&cur->task->mutex.users))
    	{
      		DQDEBUG("Deleting %d's queue entry\n", cur->task->pid);
      		SrrTask_delete(cur->task, map, 1);
    
      		if(cur->parent != NULL)
        		cur->parent->next = cur->next;

      		if(cur->next != NULL)
        		cur->next->parent = cur->parent;
      
      		if(map->deleteQueue->head == cur)
        		map->deleteQueue->head = cur->next;
      
      		if(map->deleteQueue->tail == cur)
        		map->deleteQueue->tail = cur->next;

      		xxfree(cur);
    	}
    	else if(!cur->task)
    	{
      		if(cur->parent != NULL)
        		cur->parent->next = cur->next;

      		if(cur->next != NULL)
        		cur->next->parent = cur->parent;
      
      		if(map->deleteQueue->head == cur)
        		map->deleteQueue->head = cur->next;
      
      		if(map->deleteQueue->tail == cur)
        		map->deleteQueue->tail = cur->next;
      
      		xxfree(cur);
    	}
    	else 
    	{
      		missed++;
    	}
    	cur = next;
  	}
  	DQDEBUG("Missed %d tasks in delete\n", missed);
  	mutex_unlock(&map->deleteQueue->lock);
  	return missed;
}
#endif
Static SrrTask* SrrTm_GetLocked(SrrTm* map, pid_t pid)
{
	/* HASH ACCESS */
	int i = pid;
  	int accessCount = 0;

	/* LINEAR ACCESS
    int i;
	 */
	/* LINEAR ACCESS METHOD
	for(i = 0; i < NR_TASKS; ++i) {
		if(map->tasks[i] && map->tasks[i]->pid == pid)
			return map->tasks[i];
	}
	*/
  	RemoveMapWriter();  
  	AddMapReader();
  	while(1)
	{
		if(map->tasks[i % NR_TASKS])
		{
			if(map->tasks[i % NR_TASKS]->pid == pid)
			{
        		RemoveMapReader();
        		AddMapWriter(); 
        		return map->tasks[i % NR_TASKS];
			}
		}
		if(accessCount == NR_TASKS)
		{
      		RemoveMapReader();
      		AddMapWriter();
			return 0;
		}
		accessCount++;
		i++;
	}
  	RemoveMapReader();
  	AddMapWriter();
	return 0;
}
#ifdef PRINT_TASK_TABLE
void printTm(SrrTm *map)
{
  	int i;
  	printk(KERN_DEBUG "Printing task table on behalf of %d\n", current->pid);
  	printk(KERN_DEBUG "    HASH       ADDRESS       PID       USERS\n");
  	printk(KERN_DEBUG "------------------------------------------------\n");
  	for(i = 0; i < NR_TASKS; i++)
  	{
    	if(map->tasks[i])
      		printk("    %d        %x        %d        %d\n", i, (int)map->tasks[i], map->tasks[i]->pid, atomic_read(&map->tasks[i]->mutex.users));
  	}
}
#endif
